#include <iostream>
#include <sstream>

#include "dbms.h"
#include "config.h"
#include "datagen.h"
#include "timer.h"

#include <stdlib.h>

using namespace std;

typedef struct
{
  int id;
  long long datetime_ms;
} ana_update_info_t;

typedef struct
{
  int id;
  long long datetime_ms;
  bool value;
} bool_update_info_t;

typedef struct
{
  int update_nb;
  void * pos;
}update_tab_t;

int compare(const void * a, const void * b)
{
  return ((update_tab_t*)a)->update_nb - ((update_tab_t*)b)->update_nb;
}

static uint64_t cb2_1count = 0;
void cb2_1(uint32_t anaid, uint64_t date, float value,
    unsigned char quality)
{
  cb2_1count++;
}
void cb2_1end()
{
  cout << "R2.1: nb_rows = " << cb2_1count << "\n";
  cb2_1count = 0;
}

static uint64_t cb2_2count = 0;
void cb2_2(uint32_t boolid, uint64_t date, bool value,
    unsigned char quality)
{
  cb2_2count++;
}
void cb2_2end()
{
  cout << "R2.2: nb_rows = " << cb2_2count << "\n";
  cb2_2count = 0;
}

void cb3_1(uint64_t count)
{
  cout << "R3.1: count(*) = " << count << "\n";
}

void cb3_2(uint64_t count)
{
  cout << "R3.2: count(*) = " << count << "\n";
}

void cb4(double sum)
{
  cout << "R4: sum(value) = " << sum << "\n";
}

void cb5(double avg)
{
  cout << "R5: avg(value) = " << avg << "\n";
}

void cb6(float min, float max)
{
  cout << "R6: min(value) = " << min << ", max(value) = " <<  max << "\n";
}

static uint64_t cb7count = 0;
void cb7(uint64_t date, float value)
{
  cb7count++;
}
void cb7end()
{
  cout << "R7: nb_rows = " << cb7count << "\n";
  cb7count = 0;
}

static uint64_t cb8count = 0;
void cb8(uint64_t date, float value)
{
  cb8count++;
}
void cb8end()
{
  cout << "R8: nb_rows = " << cb8count << "\n";
  cb8count = 0;
}

void cb9(const char * label, uint64_t count)
{
  cout << "R9: label = \"" << label << "\", count(*) = " << count << "\n";
}

void cb10(uint32_t anaid, uint64_t count)
{
  cout << "R10: id = " << anaid << ", count(*) = " << count << "\n";
}

void cb11_1(uint32_t anaid, float value)
{
  cout << "R11.1: id = " << anaid << ", value = " << value << "\n";
}

void cb11_2(uint32_t boolid, bool value)
{
  cout << "R11.2: id = " << boolid << ", value = " << value << "\n";
}

std::string curdate()
{
  time_t rawtime;
  struct tm * timeinfo;
  char tbuff [80];
  time (&rawtime);
  timeinfo = localtime (&rawtime);
  strftime (tbuff,80,"%Y-%m-%d %H:%M:%S",timeinfo);
  std::stringstream tmpsstr;
  /* tmpsstr << tbuff; */
  tmpsstr << rawtime;
  return(tmpsstr.str());
}

int main(int argc, char **argv)
{
  timer t;
  t.tick();

  /* display config */
  time_t rawtime;
  struct tm * timeinfo;
  char tbuff [80];
  time (&rawtime);
  timeinfo = localtime (&rawtime);
  strftime (tbuff,80,"%Y-%m-%d %H:%M:%S",timeinfo);
  cout << "date: " << tbuff << endl;
  cout << "dbms: " << dbms::name() << endl;
  cout << "version: " << dbms::version() << endl;
  cout << "filesystem: " << dbms::filesystem() << endl;
  cout << endl;
  cout << "ANA_TAGS: " << CFG_ANA_TAGS << endl;
  cout << "TOR_TAGS: " << CFG_TOR_TAGS << endl;
  cout << "R0.1_COUNT: " << CFG_R0_1_CNT << endl;
  cout << "R0.2_COUNT: " << CFG_R0_2_CNT << endl;
  cout << "R1.1_COUNT: " << CFG_R1_1_CNT << endl;
  cout << "R1.2_COUNT: " << CFG_R1_2_CNT << endl;
  cout << "R2.1_COUNT: " << CFG_R2_1_CNT << endl;
  cout << "R2.2_COUNT: " << CFG_R2_2_CNT << endl;
  cout << "R3.1_COUNT: " << CFG_R3_1_CNT << endl;
  cout << "R3.2_COUNT: " << CFG_R3_2_CNT << endl;
  cout << "R4_COUNT: " << CFG_R4_CNT << endl;
  cout << "R5_COUNT: " << CFG_R5_CNT << endl;
  cout << "R6_COUNT: " << CFG_R6_CNT << endl;
  cout << "R7_COUNT: " << CFG_R7_CNT << endl;
  cout << "R8_COUNT: " << CFG_R8_CNT << endl;
  cout << "R9_COUNT: " << CFG_R9_CNT << endl;
  cout << "R10_COUNT: " << CFG_R10_CNT << endl;
  cout << "R11.1_COUNT: " << CFG_R11_1_CNT << endl;
  cout << "R11.2_COUNT: " << CFG_R11_2_CNT << endl;
  cout << "SELECT_AVG_POINTS: " << CFG_SELECT_CNT << endl;
  cout << "SELECT_ONLY: " << CFG_SELECT_ONLY << endl;
  cout << "SEED: " << CFG_SEED << endl;
  rawtime = CFG_STARTDATE/1000;
  timeinfo = gmtime(&rawtime);
  strftime (tbuff,80,"%Y-%m-%d %H:%M:%S.",timeinfo);
  cout << "STARTDATE: " << CFG_STARTDATE
    << " (" << tbuff << CFG_STARTDATE%1000 << ")" << endl;
  cout << endl;
  cout << "dbms-specific options:" << endl;
  dbms::print_options();
  cout << endl;

  cout << "Benchmark start: " << curdate() << endl;
  dbms dbms;
  
  //init seed values-----------------------------------------------
  int initseed = CFG_SEED;
  int anaseed = CFG_SEED+1;
  int boolseed = CFG_SEED+2;
  int queriesseed = CFG_SEED+3;

  //init updates values--------------------------------------------
  /* generate random ana updates */
  CRandomMother random(initseed);
  update_tab_t * anaupdates;
  anaupdates = (update_tab_t*)malloc(CFG_R1_1_CNT*sizeof(update_tab_t));
  if(anaupdates==NULL)
  {
    cerr << "erreur d'allocation memoire anaupdates" << endl;
    exit(-1);
  }
  ana_update_info_t * anaupdates_infos;
  anaupdates_infos = (ana_update_info_t*)malloc(CFG_R1_1_CNT*sizeof(ana_update_info_t));
  if(anaupdates_infos==NULL)
  {
    cerr << "erreur d'allocation memoire anaupdates_infos" << endl;
    exit(-1);
  }
  for(int i=0; i<CFG_R1_1_CNT; i++)
  {
    anaupdates[i].update_nb = random.IRandom(0, CFG_R0_1_CNT-1);
    anaupdates[i].pos = &(anaupdates_infos[i]);
  }
  /* sort to efficiently add data */
  qsort(anaupdates, CFG_R1_1_CNT, sizeof(update_tab_t), compare);

  /* generate random bool updates */
  update_tab_t * boolupdates;
  boolupdates = (update_tab_t*)malloc(CFG_R1_2_CNT*sizeof(update_tab_t));
  if(boolupdates==NULL)
  {
    cerr << "erreur d'allocation memoire boolupdates" << endl;
    exit(-1);
  }
  bool_update_info_t * boolupdates_infos;
  boolupdates_infos = (bool_update_info_t*)malloc(CFG_R1_2_CNT*sizeof(bool_update_info_t));
  if(boolupdates_infos==NULL)
  {
    cerr << "erreur d'allocation memoire boolupdates_infos" << endl;
    exit(-1);
  }
  for(int i=0; i<CFG_R1_2_CNT; i++)
  {
    boolupdates[i].update_nb = random.IRandom(0, CFG_R0_2_CNT-1);
    boolupdates[i].pos = &(boolupdates_infos[i]);
  }
  /* sort to efficiently add data */
  qsort(boolupdates, CFG_R1_2_CNT, sizeof(update_tab_t), compare);

  /* add ana data */
  datagen anadatagen(anaseed);
  int cur=0;
  if(CFG_R1_1_CNT > 0)
  {
    for(int i=0; i<CFG_R0_1_CNT; i++)
    {
      uint32_t anaid = anadatagen.getAnaId();
      uint64_t date = anadatagen.getDate();
      anadatagen.getAnaValue();
      anadatagen.getQuality();
      while(cur<CFG_R1_1_CNT && anaupdates[cur].update_nb==i)
      {
        ((ana_update_info_t*)anaupdates[cur].pos)->id = anaid;
        ((ana_update_info_t*)anaupdates[cur].pos)->datetime_ms = date;
        cur++;
      }
    }
  }
  free(anaupdates);

  /* add bool data */
  datagen booldatagen(boolseed);
  cur=0;
  if(CFG_R1_2_CNT > 0)
  {
    for(int i=0; i<CFG_R0_2_CNT; i++)
    {
      uint32_t boolid = booldatagen.getBoolId();
      uint64_t date = booldatagen.getDate();
      bool value = booldatagen.getBoolValue();
      booldatagen.getQuality();
      while(cur<CFG_R1_2_CNT && boolupdates[cur].update_nb==i)
      {
        ((bool_update_info_t*)boolupdates[cur].pos)->id = boolid;
        ((bool_update_info_t*)boolupdates[cur].pos)->datetime_ms = date;
        ((bool_update_info_t*)boolupdates[cur].pos)->value = !(value);
        cur++;
      }
    }
  }
  free(boolupdates);

  //init dbms------------------------------------------------------
  datagen initdatagen(initseed);


  //start timer----------------------------------------------------
  cout << "Benchmark initialization OK (" << (float)t.tick()/1000000 << "sec)" << endl;
  t.tick(); /* refresh timer */

  //init database--------------------------------------------------
  for(int i=1; i<=CFG_ANA_TAGS; i++)
  {
    char label[CFG_LABEL_LEN + 1];
    initdatagen.getLabel(label);
    dbms.insertAna(
        i,                              /* id               */
        label,                          /* label            */
        initdatagen.getCreationDate(),  /* creation date    */
        NULL,                           /* destruction date */
        initdatagen.getUnit(),          /* unit             */
        initdatagen.getInterval(),      /* period           */
        initdatagen.getThreshold1(),    /* threshold1       */
        initdatagen.getThreshold2()     /* threshold2       */
        );
  }

  initdatagen.reset();

  for(int i=1; i<=CFG_TOR_TAGS; i++)
  {
    char label[CFG_LABEL_LEN + 1];
    initdatagen.getLabel(label);
    char label0[CFG_LABEL_LEN + 1];
    initdatagen.getThresholdLabel(label0);
    char label1[CFG_LABEL_LEN + 1];
    initdatagen.getThresholdLabel(label1);
    dbms.insertBool(
        i,                              /* id               */
        label,                          /* label            */
        initdatagen.getCreationDate(),  /* creation date    */
        NULL,                           /* destruction date */
        label0,                         /* label0           */
        label1                          /* label1           */
        );
  }
  //get init duration----------------------------------------------
  cout << "init duration : " << (float)t.tick()/1000000 << "sec" << endl;

  /* reset ana datagen */
  anadatagen.reset();
  t.tick(); /* refresh timer */
  cout << "R0.1 start: " << curdate() << endl;

  //insert anavalues ----------------------------------------------
  {
    uint32_t anaid;
    uint64_t date;
    float value;
    unsigned char quality;
    for(int i=0; i < CFG_R0_1_CNT; i++)
    {
      anaid = anadatagen.getAnaId();
      date = anadatagen.getDate();
      value = anadatagen.getAnaValue();
      quality = anadatagen.getQuality();
      dbms.queryR0_1(anaid, date, value, quality);
    }
  }

  //get ana insert duration----------------------------------------
  cout << "R0.1 end: " << curdate() << endl;
  cout << "R0.1 duration : " << (float)t.tick()/1000000 << "sec" << endl;

  //wait for delayed ana insert to finish--------------------------
  dbms.flushR0_1();

  /* reset bool datagen */
  booldatagen.reset();
  t.tick(); /* refresh timer */
  cout << "R0.2 start: " << curdate() << endl;

  //insert boolvalues ---------------------------------------------
  {
    uint32_t boolid;
    uint64_t date;
    bool value;
    unsigned char quality;
    for(int i=0; i < CFG_R0_2_CNT; i++)
    {
      boolid = booldatagen.getBoolId();
      date = booldatagen.getDate();
      value = booldatagen.getBoolValue();
      quality = booldatagen.getQuality();
      dbms.queryR0_2(boolid, date, value, quality);
    }
  }

  //get bool insert duration---------------------------------------
  cout << "R0.2 end: " << curdate() << endl;
  cout << "R0.2 duration : " << (float)t.tick()/1000000 << "sec" << endl;

  //wait for delayed bool insert to finish-------------------------
  dbms.flushR0_2();

  //init queries datagen-------------------------------------------
  datagen queriesdatagen(queriesseed);

  //update data----------------------------------------------------
  {//R1_1
    cout << "R1.1 start: " << curdate() << endl;
    float anavalue;
    t.tick();
    for(int i=0; i<CFG_R1_1_CNT; i++)
    {
      anavalue = queriesdatagen.getAnaValue();
      dbms.queryR1_1(anavalue, anaupdates_infos[i].id, anaupdates_infos[i].datetime_ms);
      //queriesdatagen.getAnaValue();
    }
    cout << "R1.1 end: " << curdate() << endl;
    cout << "R1.1 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }
  free(anaupdates_infos);

  {//R1_2
    cout << "R1.2 start: " << curdate() << endl;
    t.tick();
    for(int i=0; i<CFG_R1_2_CNT; i++) {
      dbms.queryR1_2(boolupdates_infos[i].value, boolupdates_infos[i].id, boolupdates_infos[i].datetime_ms);
    }
    cout << "R1.2 end: " << curdate() << endl;
    cout << "R1.2 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }
  free(boolupdates_infos);

  //queries--------------------------------------------------------
  {//R2_1
    cout << "R2.1 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R2_1_CNT; i++)
    {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR2_1(anaid, start, end, cb2_1);
      cb2_1end();
    }
    cout << "R2.1 end: " << curdate() << endl;
    cout << "R2.1 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R2_2
    cout << "R2.2 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t boolid;
    t.tick();
    for(int i=0; i<CFG_R2_2_CNT; i++)
    {
      booldatagen.getStartEnd(start, end);
      boolid = queriesdatagen.getBoolId();
      dbms.queryR2_2(boolid, start, end, cb2_2);
      cb2_2end();
    }
    cout << "R2.2 end: " << curdate() << endl;
    cout << "R2.2 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R3_1
    cout << "R3.1 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R3_1_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR3_1(anaid, start, end, cb3_1);
    }
    cout << "R3.1 end: " << curdate() << endl;
    cout << "R3.1 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R3_2
    cout << "R3.2 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t boolid;
    t.tick();
    for(int i=0; i<CFG_R3_2_CNT; i++) {
      booldatagen.getStartEnd(start, end);
      boolid = queriesdatagen.getBoolId();
      dbms.queryR3_2(boolid, start, end, cb3_2);
      //queriesdatagen.getBoolId();
    }
    cout << "R3.2 end: " << curdate() << endl;
    cout << "R3.2 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R4
    cout << "R4 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R4_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR4(anaid, start, end, cb4);
    }
    cout << "R4 end: " << curdate() << endl;
    cout << "R4 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R5
    cout << "R5 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R5_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR5(anaid, start, end, cb5);
    }
    cout << "R5 end: " << curdate() << endl;
    cout << "R5 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R6
    cout << "R6 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R6_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR6(anaid, start, end, cb6);
    }
    cout << "R6 end: " << curdate() << endl;
    cout << "R6 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R7
    cout << "R7 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    t.tick();
    for(int i=0; i<CFG_R7_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      dbms.queryR7(anaid, start, end, cb7);
      cb7end();
    }
    cout << "R7 end: " << curdate() << endl;
    cout << "R7 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R8
    cout << "R8 start: " << curdate() << endl;
    long long start;
    long long end;
    uint32_t anaid;
    float threshold;
    t.tick();
    for(int i=0; i<CFG_R8_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      anaid = queriesdatagen.getAnaId();
      threshold = queriesdatagen.getAnaValue();
      dbms.queryR8(anaid, start, end, threshold, cb8);
      cb8end();
    }
    cout << "R8 end: " << curdate() << endl;
    cout << "R8 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R9
    cout << "R9 start: " << curdate() << endl;
    long long start;
    long long end;
    t.tick();
    for(int i=0; i<CFG_R9_CNT; i++) {
      anadatagen.getStartEnd(start, end);
      dbms.queryR9(start, end, cb9);
    }
    cout << "R9 end: " << curdate() << endl;
    cout << "R9 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R10
    cout << "R10 start: " << curdate() << endl;
    long long start;
    long long end;
    t.tick();
    for(int i=0; i<CFG_R10_CNT; i++)
    {
      anadatagen.getStartEnd(start, end);
      dbms.queryR10(start, end, cb10);
    }
    cout << "R10 end: " << curdate() << endl;
    cout << "R10 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R11_1
    cout << "R11.1 start: " << curdate() << endl;
    t.tick();
    for(int i=0; i < CFG_R11_1_CNT; i++)
    {
      dbms.queryR11_1(cb11_1);
    }
    cout << "R11.1 end: " << curdate() << endl;
    cout << "R11.1 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  {//R11_2
    cout << "R11.2 start: " << curdate() << endl;
    t.tick();
    for(int i=0; i < CFG_R11_2_CNT; i++)
    {
      dbms.queryR11_2(cb11_2);
    }
    cout << "R11.2 end: " << curdate() << endl;
    cout << "R11.2 duration : " << (float)t.tick()/1000000 << "sec" << endl;
  }

  return 0;
}
