/*
* dbms.cpp
*
*  Created on: 2 sept. 2009
*      Author: chardin-bri
*/

#include "dbms.h"
#include "config.h"

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <math.h>
#include <string.h>

#include <windows.h>

#define NT
#include <setcim.h>

#define MAX_OCC 1000

#define CHK(error) do{if(error.ERRCODE != SUCCESS) { \
      ERRMESS(&error, error_msg, &errsz);         \
      error_msg[errsz] = '\0';                    \
      cerr << __LINE__ << ": " << error_msg << " - ERRCODE : " << error.ERRCODE <<endl; \
      exit(-1);                                   \
    }}while(0)                                     \

using namespace std;

static long anarecids[CFG_ANA_TAGS];
static long boolrecids[CFG_TOR_TAGS];
static long anadefid;
static long booldefid;
static long historyareaft;
static long historyarchivingft;
static long historyrepositoryft;
static long dateft;
static long valueft;
static long qualityft;
static long labelft;
static long threshold1ft;
static long threshold2ft;
static long periodft;
static long unitft;
static long creationdateft;
static long label0ft;
static long label1ft;

static ERRARRAY error_msg;
static ERRBLOCK error;
static short errsz;
static short sok;
//static unsigned long long errnb = 0;

static short anadatatypes[2];
static short booldatatypes[2];

static long fts[2];

static void * R01ptdatas[2];
static void * R02ptdatas[2];
static void * RN1ptdatas[2];
static void * RN2ptdatas[2];

static int stbool_value;
static float stana_value;
static int stquality;
static XUSTS stdate;
static float * stana_values;
static int * stbool_values;
static short * stqualities;
static XUSTS * sttimestamps;

const char * dbms::name()
{
  return "InfoPlus.21";
}

const char * dbms::version()
{
  return "2006.5";
}

const char * dbms::filesystem()
{
  return "ntfs";
}

void dbms::print_options()
{
  cout << "batch retrieval size: " << MAX_OCC << endl;
}

dbms::dbms()
{
  if(!INISETC())
  {
    cerr << "InfoPlus.21 is not up!" << endl;
    exit(-1);
  }

  string fieldname("AltValue");
  DECODFT(fieldname.c_str(), fieldname.length(), &valueft, &error);
  CHK(error);
  fieldname.assign("Quality");
  DECODFT(fieldname.c_str(), fieldname.length(), &qualityft, &error);
  CHK(error);
  fieldname.assign("AltDate");
  DECODFT(fieldname.c_str(), fieldname.length(), &dateft, &error);
  CHK(error);
  fieldname.assign("HistoryArea");
  DECODFT(fieldname.c_str(), fieldname.length(), &historyareaft, &error);
  CHK(error);
  fieldname.assign("ThresholdLow");
  DECODFT(fieldname.c_str(), fieldname.length(), &threshold1ft, &error);
  CHK(error);
  fieldname.assign("ThresholdHigh");
  DECODFT(fieldname.c_str(), fieldname.length(), &threshold2ft, &error);
  CHK(error);
  fieldname.assign("Interval");
  DECODFT(fieldname.c_str(), fieldname.length(), &periodft, &error);
  CHK(error);
  fieldname.assign("Label");
  DECODFT(fieldname.c_str(), fieldname.length(), &labelft, &error);
  CHK(error);
  fieldname.assign("CreationDate");
  DECODFT(fieldname.c_str(), fieldname.length(), &creationdateft, &error);
  CHK(error);
  fieldname.assign("Unit");
  DECODFT(fieldname.c_str(), fieldname.length(), &unitft, &error);
  CHK(error);
  fieldname.assign("Label0");
  DECODFT(fieldname.c_str(), fieldname.length(), &label0ft, &error);
  CHK(error);
  fieldname.assign("Label1");
  DECODFT(fieldname.c_str(), fieldname.length(), &label1ft, &error);
  CHK(error);
  fieldname.assign("HistoryArchiving");
  DECODFT(fieldname.c_str(), fieldname.length(), &historyarchivingft, &error);
  CHK(error);
  fieldname.assign("HistoryRepository");
  DECODFT(fieldname.c_str(), fieldname.length(), &historyrepositoryft, &error);
  CHK(error);

  string defname("Ana");
  DECODNAM(defname.c_str(), defname.length(), &anadefid, &error);
  CHK(error);
  defname.assign("Bool");
  DECODNAM(defname.c_str(), defname.length(), &booldefid, &error);
  CHK(error);

  if(CFG_SELECT_ONLY)
  {
    for(int i=0; i<CFG_ANA_TAGS; i++)
    {
      ostringstream recname;
      recname << "Ana" << (i+1);
      DECODNAM(recname.str().c_str(), recname.str().length(), &(anarecids[i]), &error);
      CHK(error);
    }

    for(int i=0; i<CFG_TOR_TAGS; i++)
    {
      ostringstream recname;
      recname << "Bool" << (i+1);
      DECODNAM(recname.str().c_str(), recname.str().length(), &(boolrecids[i]), &error);
      CHK(error);
    }
  }
  else
  {
    /* delete tags */
    unsigned int id = 1;
    do
    {
      ostringstream recname;
      recname << "Ana" << id;
      id++;
      long recid;
      DECODNAM(recname.str().c_str(), recname.str().length(), &recid, &error);
      if(error.ERRCODE == SUCCESS)
      {
        MAKUSABL(recid, &error);
        CHK(error);
        SHRT2DB(recid, historyarchivingft, 0, &error);
        CHK(error);
        MAKUNUSA(recid, &error);
        CHK(error);
        DELETREC(recid, &error);
        CHK(error);
      }
    } while(error.ERRCODE == SUCCESS);
    id = 1;
    do
    {
      ostringstream recname;
      recname << "Bool" << id;
      id++;
      long recid;
      DECODNAM(recname.str().c_str(), recname.str().length(), &recid, &error);
      if(error.ERRCODE == SUCCESS)
      {
        MAKUSABL(recid, &error);
        CHK(error);
        SHRT2DB(recid, historyarchivingft, 0, &error);
        CHK(error);
        MAKUNUSA(recid, &error);
        CHK(error);
        DELETREC(recid, &error);
        CHK(error);
      }
    } while(error.ERRCODE == SUCCESS);
  }

  stana_values = (float *)malloc(MAX_OCC * sizeof(float));
  stbool_values = (int *)malloc(MAX_OCC * sizeof(int));
  stqualities = (short *)malloc(MAX_OCC * sizeof(short));
  sttimestamps = (XUSTS *)malloc(MAX_OCC * sizeof(XUSTS));

  fts[0] = valueft;
  fts[1] = qualityft;

  anadatatypes[0] = DTYPREAL;
  anadatatypes[1] = DTYPSHRT;
  booldatatypes[0] = DTYPSHRT;
  booldatatypes[1] = DTYPSHRT;

  R01ptdatas[0] = &(stana_value);
  R01ptdatas[1] = &(stquality);
  R02ptdatas[0] = &(stbool_value);
  R02ptdatas[1] = &(stquality);
  RN1ptdatas[0] = stana_values;
  RN1ptdatas[1] = stqualities;
  RN2ptdatas[0] = stbool_values;
  RN2ptdatas[1] = stqualities;
}

dbms::~dbms()
{
  ENDSETC();
  free(stana_values);
  free(stbool_values);
  free(stqualities);
  free(sttimestamps);
}

void dbms::insertAna(unsigned int anaid, char * label, unsigned int creationdate, unsigned int * destructiondate, unsigned int unit, unsigned int period, float threshold1, float threshold2)
{
  if(!CFG_SELECT_ONLY)
  {
    ostringstream recname;
    recname << "Ana" << anaid;
    long recid = RECIDMAX();
    do
    {
      CHKFREE(recid, &error);
      if(error.ERRCODE != SUCCESS)
      {
        recid--;
      }
    } while(error.ERRCODE != SUCCESS && recid > 0);
    CREATEREC(recid, anadefid, recname.str().c_str(), recname.str().length(), &error);
    CHK(error);
    anarecids[anaid - 1] = recid;

    CHBF2DB(recid, labelft, label, strlen(label), &error);
    CHK(error);
    XUSTS creatdate;
    creatdate.secs = creationdate;
    creatdate.usecs = 0;
    XTSBLOCK xts;
    XUST2XTS(&creatdate, &xts);
    XTIM2DB(recid, creationdateft, &xts, &error);
    CHK(error);
    SHRT2DB(recid, unitft, unit, &error);
    CHK(error);
    SHRT2DB(recid, periodft, period, &error);
    CHK(error);
    REALADD2DB(recid, threshold1ft, &threshold1, &error);
    CHK(error);
    REALADD2DB(recid, threshold2ft, &threshold2, &error);
    CHK(error);
    SHRT2DB(recid, historyareaft, 1, &error);
    CHK(error);
    WDBASCII(recid, historyrepositoryft, "MY_REP_ANA", 10, false, &error);
    CHK(error);
    MAKUSABL(recid, &error);
    CHK(error);
    SHRT2DB(recid, historyarchivingft, 1, &error);
    CHK(error);
  }
}

void dbms::insertBool(unsigned int boolid, char * label, unsigned int creationdate, unsigned int * destructiondate, char * label0, char * label1)
{
  if(!CFG_SELECT_ONLY)
  {
    ostringstream recname;
    recname << "Bool" << boolid;
    long recid = RECIDMAX();
    do
    {
      CHKFREE(recid, &error);
      if(error.ERRCODE != SUCCESS)
      {
        recid--;
      }
    } while(error.ERRCODE != SUCCESS && recid > 0);
    CREATEREC(recid, booldefid, recname.str().c_str(), recname.str().length(), &error);
    CHK(error);
    boolrecids[boolid - 1] = recid;

    CHBF2DB(recid, labelft, label, strlen(label), &error);
    CHK(error);
    XUSTS creatdate;
    creatdate.secs = creationdate;
    creatdate.usecs = 0;
    XTSBLOCK xts;
    XUST2XTS(&creatdate, &xts);
    XTIM2DB(recid, creationdateft, &xts, &error);
    CHK(error);
    CHBF2DB(recid, label0ft, label0, strlen(label0), &error);
    CHK(error);
    CHBF2DB(recid, label1ft, label1, strlen(label1), &error);
    CHK(error);
    SHRT2DB(recid, historyareaft, 1, &error);
    CHK(error);
    WDBASCII(recid, historyrepositoryft, "MY_REP_TOR", 10, false, &error);
    CHK(error);
    MAKUSABL(recid, &error);
    CHK(error);
    SHRT2DB(recid, historyarchivingft, 1, &error);
    CHK(error);
  }
}

void dbms::flushR0_1()
{

}

void dbms::flushR0_2()
{

}

void dbms::queryR0_1(unsigned int anaid, unsigned long long date, float ana_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    stana_value = ana_value;
    stquality = quality;
    stdate.secs = date/1000;
    stdate.usecs = (date%1000)*1000;

    do
    {
      WHIS21DAT(WHIS_TYPE_ADDNEW, anarecids[anaid - 1],
      historyareaft, 2, fts, anadatatypes, R01ptdatas, -1, &stdate,
      &sok, &error);
      if(error.ERRCODE != SUCCESS)
      {
        if(error.ERRCODE == DHNOREAD) //wait if database is not ready
        {
          Sleep(1000);
          //errnb++;
        }
        else //else display error
        {
          ERRMESS(&error, error_msg, &errsz);
          error_msg[errsz] = '\0';
          cerr << "R0.1: " << error_msg << " - ERRCODE : " << error.ERRCODE <<endl;
          if(error.ERRCODE == INVKEYTS) //stop if invalid key (date)
          {
            cerr << "invalid date is : " << date << endl;
            exit(-1);
          }
          else if(error.ERRCODE == DHWRITER) //ignore this error
          {
            error.ERRCODE = SUCCESS;
          }
          else
          {
            exit(-1);
          }
        }
      }
    } while(error.ERRCODE != SUCCESS);
  }
}

void dbms::queryR0_2(unsigned int boolid, unsigned long long date,
bool bool_value, unsigned char quality)
{
  if(!CFG_SELECT_ONLY)
  {
    stbool_value = bool_value;
    stquality = quality;
    stdate.secs = date/1000;
    stdate.usecs = (date%1000)*1000;

    do
    {
      WHIS21DAT(WHIS_TYPE_ADDNEW, boolrecids[boolid-1],
      historyareaft, 2, fts, booldatatypes, R02ptdatas, -1, &stdate,
      &sok, &error);
      if(error.ERRCODE != SUCCESS)
      {
        if(error.ERRCODE == DHNOREAD) //wait if database is not ready
        {
          Sleep(1000);
        }
        else //else display error
        {
          ERRMESS(&error, error_msg, &errsz);
          error_msg[errsz] = '\0';
          cerr << "R0.2: " << error_msg << " - ERRCODE : " << error.ERRCODE <<endl;
          if(error.ERRCODE == INVKEYTS) //stop if invalid key (date)
          {
            cerr << "invalid date is : " << date << endl;
            exit(-1);
          }
          else if(error.ERRCODE == DHWRITER) //ignore this error
          {
            error.ERRCODE = SUCCESS;
          }
          else
          {
            exit(-1);
          }
        }
      }
    } while(error.ERRCODE != SUCCESS);
  }
}

void dbms::queryR1_1(float Value, unsigned int AnaId, unsigned long long Date)
{
  if(!CFG_SELECT_ONLY)
  {
    stana_value = Value;
    stdate.secs = Date/1000;
    stdate.usecs = (Date%1000)*1000;

    WHIS21DAT(WHIS_TYPE_MODIFY, anarecids[AnaId - 1], historyareaft, 2, fts,
    anadatatypes, R01ptdatas, -1, &stdate, &sok, &error);
    if(error.ERRCODE != SUCCESS)
    {
      ERRMESS(&error, error_msg, &errsz);
      error_msg[errsz] = '\0';
      cerr << __LINE__ << ": " << error_msg << " - ERRCODE : " << error.ERRCODE <<endl;
      if(error.ERRCODE == INVKEYTS)
      {
        cerr << "invalid (id, date) is : (" << AnaId << ", " << Date << ")" << endl;
      }
    }
  }
}

void dbms::queryR1_2(bool Value, unsigned int BoolId,
unsigned long long Date)
{
  if(!CFG_SELECT_ONLY)
  {
    if(Value)
    {
      stbool_value = 1;
    }
    else
    {
      stbool_value = 0;
    }
    stdate.secs = Date/1000;
    stdate.usecs = (Date%1000)*1000;

    WHIS21DAT(WHIS_TYPE_MODIFY, boolrecids[BoolId - 1], historyareaft, 2, fts,
    booldatatypes, R02ptdatas, -1, &stdate, &sok, &error);
    if(error.ERRCODE != SUCCESS)
    {
      ERRMESS(&error, error_msg, &errsz);
      error_msg[errsz] = '\0';
      cerr << __LINE__ << ": " << error_msg << " - ERRCODE : " << error.ERRCODE <<endl;
      if(error.ERRCODE == INVKEYTS)
      {
        cerr << "invalid (id, date) is : (" << BoolId << ", " << Date << ")" << endl;
      }
    }
  }
}

void dbms::queryR2_1(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)
(unsigned int, unsigned long long, float, unsigned char))
{
  //ostringstream query;
  //query << "SELECT * FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;

  long occok = MAX_OCC;
  short ftsok = 0;
  unsigned long long retdate;

  while(occok == MAX_OCC)
  {
    RHIS21DATA(H21_GET_ACTUALS, 0, 0, anarecids[AnaId - 1], historyareaft,
    &xstart, &xend, 2, fts, anadatatypes, MAX_OCC, NULL, sttimestamps,
    RN1ptdatas, &occok, &ftsok, &error);
    CHK(error);
    xstart.secs = sttimestamps[occok - 1].secs;
    xstart.usecs = sttimestamps[occok - 1].usecs + 1;
    if(xstart.usecs >= 1000000)
    {
      xstart.usecs-=1000000;
      xstart.secs++;
    }
    for(int i=0; i < occok; i++)
    {
      retdate = sttimestamps[i].secs * 1000 + sttimestamps[i].usecs/1000;
      callback(AnaId, retdate, stana_values[i], stqualities[i]);
    }
  }
}

void dbms::queryR2_2(unsigned int BoolId, unsigned long long start,
unsigned long long end, void (*callback)
(unsigned int, unsigned long long, bool, unsigned char))
{
  //ostringstream query;
  //query << "SELECT * FROM BoolValues WHERE BoolId=" << BoolId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;

  long occok = MAX_OCC;
  short ftsok = 0;
  unsigned long long retdate;

  while(occok == MAX_OCC)
  {
    RHIS21DATA(H21_GET_ACTUALS, 0, 0, boolrecids[BoolId - 1], historyareaft,
    &xstart, &xend, 2, fts, booldatatypes, MAX_OCC, NULL, sttimestamps,
    RN2ptdatas, &occok, &ftsok, &error);
    CHK(error);
    xstart.secs = sttimestamps[occok - 1].secs;
    xstart.usecs = sttimestamps[occok - 1].usecs + 1;
    if(xstart.usecs >= 1000000)
    {
      xstart.usecs-=1000000;
      xstart.secs++;
    }
    for(int i=0; i < occok; i++)
    {
      retdate = sttimestamps[i].secs * 1000 + sttimestamps[i].usecs/1000;
      callback(BoolId, retdate, stbool_values[i], stqualities[i]);
    }
  }
}

void dbms::queryR3_1(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)(unsigned long long))
{
  //ostringstream query;
  //query << "SELECT count(*) FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;
  XUSTS interval;
  interval.secs = xend.secs - xstart.secs + 1;
  interval.usecs = 0;

  double count;
  long numperiods;
  short doublecodes[1] = {AG_DBL_GOOD};
  
  RHIS21AGGREG(
      0,            //timeweight
      0,            //step
      anarecids[AnaId - 1],        //recid
      valueft | 0x00000001,   //ft
      &xstart,       //ptTimeOld
      &xend,         //ptTimeNew
      &interval,    //ptInterval
      0,            //timealign
      0,            //dsadjust
      1,            //maxperiods
      0,            //numtimecodes
      1,            //numdoublecodes
      0,            //numshortcodes
      NULL,         //timecodes
      doublecodes,  //doublecodes
      NULL,         //shortcodes
      NULL,         //timevalues
      &count,         //doublevalues
      NULL,         //shortvalues
      &numperiods,  //numperiods
      &error
  );
  CHK(error);

  callback((unsigned long long)count);
}

void dbms::queryR3_2(unsigned int BoolId, unsigned long long start,
unsigned long long end, void (*callback)(unsigned long long))
{
  //ostringstream query;
  //query << "SELECT count(*) FROM BoolValues WHERE BoolId=" << BoolId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;
  XUSTS interval;
  interval.secs = xend.secs - xstart.secs + 1;
  interval.usecs = 0;

  double count;
  long numperiods;
  short doublecodes[1] = {AG_DBL_GOOD};
  
  RHIS21AGGREG(
      0,            //timeweight
      0,            //step
      boolrecids[BoolId - 1],        //recid
      valueft | 0x00000001,   //ft
      &xstart,       //ptTimeOld
      &xend,         //ptTimeNew
      &interval,    //ptInterval
      0,            //timealign
      0,            //dsadjust
      1,            //maxperiods
      0,            //numtimecodes
      1,            //numdoublecodes
      0,            //numshortcodes
      NULL,         //timecodes
      doublecodes,  //doublecodes
      NULL,         //shortcodes
      NULL,         //timevalues
      &count,         //doublevalues
      NULL,         //shortvalues
      &numperiods,  //numperiods
      &error
  );
  CHK(error);

  callback((unsigned long long)count);
}

void dbms::queryR4(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)(double))
{
  //ostringstream query;
  //query << "SELECT sum(Value) FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;
  XUSTS interval;
  interval.secs = xend.secs - xstart.secs + 1;
  interval.usecs = 0;

  double sum;
  long numperiods;
  short doublecodes[1] = {AG_DBL_SUM};
  
  RHIS21AGGREG(
      0,            //timeweight
      0,            //step
      anarecids[AnaId - 1],        //recid
      valueft | 0x00000001,   //ft
      &xstart,       //ptTimeOld
      &xend,         //ptTimeNew
      &interval,    //ptInterval
      0,            //timealign
      0,            //dsadjust
      1,            //maxperiods
      0,            //numtimecodes
      1,            //numdoublecodes
      0,            //numshortcodes
      NULL,         //timecodes
      doublecodes,  //doublecodes
      NULL,         //shortcodes
      NULL,         //timevalues
      &sum,         //doublevalues
      NULL,         //shortvalues
      &numperiods,  //numperiods
      &error
  );
  CHK(error);

  callback(sum);
}

void dbms::queryR5(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)(double))
{
  //ostringstream query;
  //query << "SELECT avg(Value) FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;
  XUSTS interval;
  interval.secs = xend.secs - xstart.secs + 1;
  interval.usecs = 0;

  double avg;
  long numperiods;
  short doublecodes[1] = {AG_DBL_AVG};
  
  RHIS21AGGREG(
      0,            //timeweight
      0,            //step
      anarecids[AnaId - 1],        //recid
      valueft | 0x00000001,   //ft
      &xstart,       //ptTimeOld
      &xend,         //ptTimeNew
      &interval,    //ptInterval
      0,            //timealign
      0,            //dsadjust
      1,            //maxperiods
      0,            //numtimecodes
      1,            //numdoublecodes
      0,            //numshortcodes
      NULL,         //timecodes
      doublecodes,  //doublecodes
      NULL,         //shortcodes
      NULL,         //timevalues
      &avg,         //doublevalues
      NULL,         //shortvalues
      &numperiods,  //numperiods
      &error
  );
  CHK(error);
 
  callback(avg);
}

void dbms::queryR6(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)(float, float)) {
  //ostringstream query;
  //query << "SELECT min(Value), max(Value) FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;
  XUSTS interval;
  interval.secs = xend.secs - xstart.secs + 1;
  interval.usecs = 0;

  double minmax[2];
  long numperiods;
  short doublecodes[2] = {AG_DBL_MIN, AG_DBL_MAX};
  
  RHIS21AGGREG(
      0,            //timeweight
      0,            //step
      anarecids[AnaId - 1],        //recid
      valueft | 0x00000001,   //ft
      &xstart,       //ptTimeOld
      &xend,         //ptTimeNew
      &interval,    //ptInterval
      0,            //timealign
      0,            //dsadjust
      1,            //maxperiods
      0,            //numtimecodes
      2,            //numdoublecodes
      0,            //numshortcodes
      NULL,         //timecodes
      doublecodes,  //doublecodes
      NULL,         //shortcodes
      NULL,         //timevalues
      minmax,         //doublevalues
      NULL,         //shortvalues
      &numperiods,  //numperiods
      &error
  );
  CHK(error);

  callback(minmax[0], minmax[1]);
}

void dbms::queryR7(unsigned int AnaId, unsigned long long start,
unsigned long long end, void (*callback)(unsigned long long, float))
{
  //ostringstream query;
  //query << "SELECT Date, Value FROM Ana, AnaValues WHERE Ana.AnaId = AnaValues.AnaId AND AnaValues.AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end << " AND Value>Ana.Threshold2";

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;

  long occok = MAX_OCC;
  short ftsok = 0;

  float threshold2;

  DB2REAL(anarecids[AnaId - 1], threshold2ft, &threshold2, &error);
  CHK(error);

  while(occok == MAX_OCC)
  {
    RHIS21DATA (H21_GET_ACTUALS, 0, 0, anarecids[AnaId - 1], historyareaft,
    &xstart, &xend, 2, fts, anadatatypes, MAX_OCC, NULL, sttimestamps, RN1ptdatas, &occok, &ftsok,
    &error);
    CHK(error);
    xstart.secs = sttimestamps[occok - 1].secs;
    xstart.usecs = sttimestamps[occok - 1].usecs + 1;
    if(xstart.usecs >= 1000000)
    {
      xstart.usecs-=1000000;
      xstart.secs++;
    }
    for(int i=0; i<occok; i++)
    {
      if(stana_values[i] > threshold2)
      {
        callback(((unsigned long long)sttimestamps[i].secs) * 1000 + sttimestamps[i].usecs / 1000, stana_values[i]);
      }
    }
  }
}

void dbms::queryR8(unsigned int AnaId, unsigned long long start,
unsigned long long end, float Threshold, void (*callback)
(unsigned long long, float))
{
  //ostringstream query;
  //query << "SELECT Date, Value FROM AnaValues WHERE AnaId=" << AnaId << " AND Date BETWEEN " << start << " AND " << end << " AND Value>" << Threshold;

  XUSTS xstart;
  xstart.secs = start/1000;
  xstart.usecs = (start%1000)*1000;
  XUSTS xend;
  xend.secs = end/1000;
  xend.usecs = (end%1000)*1000;

  long occok = MAX_OCC;
  short ftsok = 0;

  while(occok == MAX_OCC)
  {
    RHIS21DATA (H21_GET_ACTUALS, 0, 0, anarecids[AnaId - 1], historyareaft,
    &xstart, &xend, 2, fts, anadatatypes, MAX_OCC, NULL, sttimestamps, RN1ptdatas, &occok, &ftsok,
    &error);
    CHK(error);
    xstart.secs = sttimestamps[occok - 1].secs;
    xstart.usecs = sttimestamps[occok - 1].usecs + 1;
    if(xstart.usecs >= 1000000)
    {
      xstart.usecs-=1000000;
      xstart.secs++;
    }
    for(int i=0; i<occok; i++)
    {
      if(stana_values[i] > Threshold)
      {
        callback(((unsigned long long)sttimestamps[i].secs) * 1000 + sttimestamps[i].usecs / 1000, stana_values[i]);
      }
    }
  }
}

/*
typedef struct tmpR9_st
{
  char label[80];
  unsigned int anaid;
  float threshold1;
  float threshold2;
} tmpR9_t;
*/

void dbms::queryR9(unsigned long long start, unsigned long long end,
void (*callback)(const char *, unsigned long long))
{
  /*
  query << "SELECT Label, count(*) as count FROM Ana STRAIGHT JOIN AnaValues FORCE INDEX (PRIMARY) ON Ana.AnaId=AnaValues.AnaId WHERE Date BETWEEN " << start << " AND " << end << " AND (Value>Threshold2 OR Value<Threshold1) GROUP BY Ana.AnaId ORDER BY count DESC LIMIT 1";
  */

  char labelmax[81] = "";
  unsigned long long maxcount = 0;

  for(int j=0; j < CFG_ANA_TAGS; j++)
  {

    XUSTS xstart;
    xstart.secs = start/1000;
    xstart.usecs = (start%1000)*1000;
    XUSTS xend;
    xend.secs = end/1000;
    xend.usecs = (end%1000)*1000;

    long occok = MAX_OCC;
    short ftsok = 0;

    float threshold1;
    float threshold2;
    unsigned long long count = 0;

    DB2REAL(anarecids[j], threshold1ft, &threshold1, &error);
    CHK(error);
    DB2REAL(anarecids[j], threshold2ft, &threshold2, &error);
    CHK(error);

    while(occok == MAX_OCC)
    {
      RHIS21DATA (H21_GET_ACTUALS, 0, 0, anarecids[j], historyareaft,
      &xstart, &xend, 2, fts, anadatatypes, MAX_OCC, NULL, sttimestamps, RN1ptdatas, &occok, &ftsok,
      &error);
      CHK(error);
      xstart.secs = sttimestamps[occok - 1].secs;
      xstart.usecs = sttimestamps[occok - 1].usecs + 1;
      if(xstart.usecs >= 1000000)
      {
        xstart.usecs-=1000000;
        xstart.secs++;
      }
      for(int i=0; i<occok; i++)
      {
        if(stana_values[i] < threshold1 || stana_values[i] > threshold2)
        {
          count++;
        }
      }
    }

    if(count > maxcount)
    {
      maxcount = count;
      DB2CHBF(anarecids[j], labelft, labelmax, 80, &error);
      if (error.ERRCODE != SUCCESS)
      {
        ERRMESS(&error, error_msg, &errsz);
        error_msg[errsz] = '\0';
        cerr << error_msg << " - ERRCODE : " << error.ERRCODE <<endl;
        exit(-1);
      }
    }
  }

  callback(labelmax, maxcount);
}

void dbms::queryR10(unsigned long long start, unsigned long long end,
void (*callback)(unsigned int, unsigned long long))
{
  //ostringstream query;
  //query << "SELECT Diff.AnaId, count(*) as c, Period FROM Ana, (SELECT D1.AnaId, D1.Date, min(D2.Date-D1.Date) as diff_ms FROM AnaValues D1, AnaValues D2 WHERE D2.Date>D1.Date AND D1.AnaId=D2.AnaId AND D1.Date BETWEEN " << start << " AND " << end << " GROUP BY D1.AnaId, D1.Date) as Diff WHERE Diff.AnaId = Ana.AnaId AND Diff.diff_ms>Ana.Period GROUP BY Diff.AnaId ORDER BY c DESC LIMIT 1";

  unsigned int idmax = 0;
  unsigned long long maxcount = 0;

  for(int j=0; j < CFG_ANA_TAGS; j++)
  {

    XUSTS xstart;
    xstart.secs = start/1000;
    xstart.usecs = (start%1000)*1000;
    XUSTS xend;
    xend.secs = end/1000;
    xend.usecs = (end%1000)*1000;
    unsigned long long prevdate = 0;
    unsigned long long date = 0;

    long occok = MAX_OCC;
    short ftsok = 0;

    long period;
    unsigned long long count = 0;

    DB2LONG(anarecids[j], periodft, &period, &error);
    CHK(error);
    period = period * 1000;

    while(occok == MAX_OCC)
    {
      RHIS21DATA (H21_GET_ACTUALS, 0, 0, anarecids[j], historyareaft,
      &xstart, &xend, 2, fts, anadatatypes, MAX_OCC, NULL, sttimestamps, RN1ptdatas, &occok, &ftsok,
      &error);
      CHK(error);
      xstart.secs = sttimestamps[occok - 1].secs;
      xstart.usecs = sttimestamps[occok - 1].usecs + 1;
      if(xstart.usecs >= 1000000)
      {
        xstart.usecs-=1000000;
        xstart.secs++;
      }
      for(int i=0; i<occok; i++)
      {
        if(prevdate == 0)
        {
          prevdate = sttimestamps[0].secs * 1000 + sttimestamps[0].usecs/1000;
        }
        else
        {
          date = sttimestamps[i].secs * 1000 + sttimestamps[i].usecs/1000;
          if(date - prevdate > period)
          {
            count++;
          }
          prevdate = date;
        }
      }
    }

    if(count > maxcount)
    {
      maxcount = count;
      idmax = j+1;
    }
  }

  callback(idmax, maxcount);
}

void dbms::queryR11_1(void (*callback)(unsigned int, float))
{
  //ostringstream query;
  //query << "SELECT AnaValues.AnaId, Value FROM AnaValues, (SELECT AnaId, max(Date) as maxdate FROM AnaValues GROUP BY AnaId) as maxdates WHERE Date=maxdates.maxdate AND AnaValues.AnaId = maxdates.AnaId ORDER BY AnaValues.AnaId";

  for(int j=0; j < CFG_ANA_TAGS; j++)
  {
    short occok = 0;
    short ftsok = 0;

    MRDBOCCS(anarecids[j], 2, fts, anadatatypes, 1, 1, RN1ptdatas, &occok, &ftsok, &error);
    CHK(error);

    callback(j+1, stana_values[0]);
  }
}

void dbms::queryR11_2(void (*callback)(unsigned int, bool))
{
  //ostringstream query;
  //query << "SELECT BoolValues.BoolId, Value FROM BoolValues FORCE INDEX (PRIMARY), (SELECT BoolId, max(Date) as maxdate FROM BoolValues FORCE INDEX (PRIMARY) GROUP BY BoolId) as maxdates WHERE Date=maxdates.maxdate AND BoolValues.BoolId = maxdates.BoolId ORDER BY BoolValues.BoolId";

  for(int j=0; j < CFG_TOR_TAGS; j++)
  {
    short occok = 0;
    short ftsok = 0;

    MRDBOCCS(boolrecids[j], 2, fts, booldatatypes, 1, 1, RN2ptdatas, &occok, &ftsok, &error);
    CHK(error);

    callback(j+1, stbool_values[0]);
  }
}

