#ifndef TIMER_H_
#define TIMER_H_

#ifndef _WIN32
#include <time.h>
#include <sys/time.h>
#endif

/**
 * Timer to get time differences between tick() calls.
 */
class timer {
private :
#ifdef _WIN32
  long long prevTime;           // previous performance counter value
  long long frequency;          //ticks per sec
#else
  unsigned long long prevTime;  // previous performance counter value
  float ns_per_tick;            // nanoseconds per performance counter
  struct timespec prevTimespec;
  struct timespec res;
  struct timeval tp;
#endif
public:
  timer();
  ~timer();
  /**
   * Get time elapsed since last tick() call, in microseconds.
   * The first call currently returns time elapsed since program start
   * and should not be used.
   *
   * @return Time elapsed since last tick() call, in microseconds.
   */
  long long tick();
};

#endif /* TIMER_H_ */
